"""Exception handler for smart_logger.

Prints all accumulated logs and sends default error message.
"""

from typing import Awaitable, Callable

from fastapi import Request
from loguru import logger
from pybotx import Bot, IncomingMessage
from starlette.responses import Response
from starlette.status import HTTP_500_INTERNAL_SERVER_ERROR

from pybotx_smart_logger import log_levels
from pybotx_smart_logger.contextvars import get_debug_enabled
from pybotx_smart_logger.logger import flush_accumulated_logs
from pybotx_smart_logger.output import (
    attach_log_source,
    log_incoming_http_request,
    log_incoming_message,
)


def make_smart_logger_exception_handler(
    error_text: str,
) -> Callable[[IncomingMessage, Bot, Exception], Awaitable[None]]:
    async def exception_handler(
        message: IncomingMessage,
        bot: Bot,
        exc: Exception,
    ) -> None:
        if not get_debug_enabled():
            log_incoming_message(
                message.raw_command,
                "Error while processing incoming message:",
                log_levels.ERROR,
            )
            flush_accumulated_logs(log_levels.ERROR)

        logger.exception(attach_log_source(""))

        await bot.answer_message(error_text, wait_callback=False)

    return exception_handler


async def fastapi_exception_handler(request: Request, exc: Exception) -> Response:
    if not get_debug_enabled():
        flush_accumulated_logs(log_levels.ERROR)
        log_incoming_http_request(
            "Error while processing incoming http request",
            log_levels.ERROR,
        )

    logger.exception(attach_log_source(""))

    return Response(status_code=HTTP_500_INTERNAL_SERVER_ERROR)



"""Middlewares for smart logger."""

from typing import Awaitable, Callable, Union

from fastapi import Request
from pybotx import Bot, IncomingMessage, IncomingMessageHandlerFunc

from pybotx_smart_logger import log_levels
from pybotx_smart_logger.contextvars import (
    init_accumulated_logs,
    set_debug_enabled,
    set_http_request_headers,
    set_log_source,
)
from pybotx_smart_logger.output import log_incoming_http_request, log_incoming_message
from pybotx_smart_logger.schemas import LogSource


class BotXSmartLoggerMiddleware:
    def __init__(
        self,
        debug_enabled_for_message: Union[
            Callable[[IncomingMessage], Awaitable[bool]],
            bool,
        ],
    ) -> None:
        self._debug_enabled_for_message = debug_enabled_for_message

    async def dispatch(
        self,
        message: IncomingMessage,
        bot: Bot,
        call_next: IncomingMessageHandlerFunc,
    ) -> None:
        set_log_source(LogSource.from_user_message(message.sender.huid))

        if isinstance(self._debug_enabled_for_message, bool):
            debug_enabled = self._debug_enabled_for_message
        else:
            debug_enabled = await self._debug_enabled_for_message(message)

        set_debug_enabled(debug_enabled)

        if debug_enabled:
            log_incoming_message(
                message.raw_command,
                "Processing incoming_message:",
                log_levels.INFO,
            )

        await call_next(message, bot)


class FastApiSmartLoggerMiddleware:
    def __init__(
        self,
        *,
        debug_enabled: bool = False,
    ) -> None:
        self._debug_enabled = debug_enabled

    async def __call__(self, request: Request, call_next: Callable) -> None:
        # In pybotx handlers runs in a separate asyncio task.
        # If the contextvar variable is set in a child task, the parent task
        # don't know about it. Here we init the log storage with an empty list.
        # So when the child task will append it, we will find it out.
        init_accumulated_logs()

        method = request.method
        url = str(request.url)
        headers = dict(request.headers)

        set_http_request_headers(headers)
        set_log_source(LogSource.from_http_request(method, url))

        set_debug_enabled(self._debug_enabled)

        if self._debug_enabled:
            log_incoming_http_request(
                "Processing incoming http request",
                log_levels.INFO,
            )

        return await call_next(request)



"""Contextvars for smart_logger."""

from contextvars import ContextVar
from typing import Dict, Optional

from pybotx_smart_logger.schemas import AccumulatedLogs, LogSource

_accumulated_logs_var: ContextVar[Optional[AccumulatedLogs]] = ContextVar(
    "accumulated_logs",
    default=None,
)


def init_accumulated_logs() -> None:
    assert _accumulated_logs_var.get() is None
    _accumulated_logs_var.set([])


def get_accumulated_logs() -> AccumulatedLogs:
    accumulated_logs = _accumulated_logs_var.get()
    if accumulated_logs is None:
        accumulated_logs = []
        _accumulated_logs_var.set(accumulated_logs)

    return accumulated_logs


def clear_accumulated_logs() -> None:
    _accumulated_logs_var.set(None)


_debug_enabled_var: ContextVar[bool] = ContextVar("debug_enabled", default=False)


def get_debug_enabled() -> bool:
    return _debug_enabled_var.get()


def set_debug_enabled(new_value: bool) -> None:
    _debug_enabled_var.set(new_value)


_log_source_var: ContextVar[Optional[LogSource]] = ContextVar(
    "log_source_var",
    default=None,
)


def get_log_source() -> Optional[LogSource]:
    return _log_source_var.get()


def set_log_source(log_source: LogSource) -> None:
    _log_source_var.set(log_source)


Headers = Dict[str, str]
_http_request_headers: ContextVar[Optional[Headers]] = ContextVar(
    "http_request_headers",
    default=None,
)


def get_http_request_headers() -> Optional[Headers]:
    return _http_request_headers.get()


def set_http_request_headers(headers: Headers) -> None:
    _http_request_headers.set(headers)



from .contextmanager import wrap_system_event
from .exception_handler import (
    fastapi_exception_handler,
    make_smart_logger_exception_handler,
)
from .logger import smart_log
from .middlewares import BotXSmartLoggerMiddleware, FastApiSmartLoggerMiddleware
from .task_decorator import make_smart_logger_decorator
